from dataclasses import dataclass, field
from os import environ
from typing import Any, Dict, Optional
from unittest.mock import MagicMock

from microcosm.decorators import defaults
from microcosm.errors import NotBoundError
from microcosm.object_graph import ObjectGraph
from microcosm_logging.decorators import logger


try:
    import sentry_sdk
    from sentry_sdk.utils import BadDsn
except ImportError:
    sentry_sdk = None  # type:ignore


DEFAULT_TAG_MAPPING = {
    "X-Request-Id": "x-request-id",
    "message_id": "message-id",
    "media_type": "media-type",
    "span_name": "span",
    "uri": "uri",
    "handler": "handler",
}
DEFAULT_USER_ID_KEY = "X-Request-User"


@dataclass
class SentryConfigPubsub:
    dsn: Optional[str] = None
    enabled: bool = False
    client: Optional[Any] = None
    tag_mapping: Dict[str, str] = field(default_factory=dict)
    user_id_key: Optional[str] = None


def default_before_send(event: Dict[str, Any], hint) -> Dict[str, Any]:
    """
    Before sending a event to sentry scrub all values that are non-uuid data from the event.
    Allow python internals such as <method >, <func > etc. and self, cls.

    """
    for exception in event["exception"]["values"]:
        for frame in exception["stacktrace"]["frames"]:
            for var_name, value in frame["vars"].items():
                if not value:
                    continue
                if var_name in ["self", "cls"]:
                    continue
                elif isinstance(value, str):
                    if value.startswith("<") and value.endswith(">"):
                        continue
                    elif var_name.endswith("_id"):
                        continue
                    elif value.startswith("http"):
                        continue
                elif isinstance(value, dict):
                    for key in value.keys():
                        if key.endswith("_id"):
                            continue
                        value[key] = "<redacted>"
                    continue
                frame["vars"][var_name] = "<redacted>"
    return event


@defaults(
    enabled=False,
    dsn=None,
    custom_tags_mapping=None,
    custom_user_id=None,
)
@logger
def configure_sentry_pubsub(graph: ObjectGraph) -> SentryConfigPubsub:
    enabled, dsn, sentry = False, None, None
    tag_mapping, custom_user_id_key = DEFAULT_TAG_MAPPING, DEFAULT_USER_ID_KEY

    if graph.config.sentry_logging_pubsub.enabled and graph.config.sentry_logging_pubsub.dsn and sentry_sdk:
        enabled = graph.config.sentry_logging_pubsub.enabled
        dsn = graph.config.sentry_logging_pubsub.dsn

        try:
            before_send_func = graph.sentry_before_send
        except NotBoundError:
            before_send_func = default_before_send

        build_info = getattr(graph.config, "build_info_convention", None)
        sentry_kwargs = dict(
            dsn=dsn,
            server_name=f"{graph.metadata.name}_daemon",
            release=getattr(build_info, "sha1", None),
            environment=environ.get("MICROCOSM_ENVIRONMENT") or "undefined",
            before_send=before_send_func,
        )

        if graph.metadata.testing:
            sentry = MagicMock(**sentry_kwargs)
        else:
            try:
                sentry = sentry_sdk.init(**sentry_kwargs)  # type: ignore
            except BadDsn:
                enabled, dsn, sentry = False, None, None
                configure_sentry_pubsub.logger.error("Bad DSN value set")
    if graph.config.sentry_logging_pubsub.custom_tags_mapping:
        tag_mapping.update(graph.config.sentry_logging_pubsub.custom_tags)
    if graph.config.sentry_logging_pubsub.custom_user_id:
        custom_user_id_key = graph.config.sentry_logging_pubsub.custom_user_id

    return SentryConfigPubsub(
        dsn=dsn,
        enabled=enabled,
        client=sentry,
        tag_mapping=tag_mapping,
        user_id_key=custom_user_id_key,
    )



"""
Constants.

"""

DEFAULT_RESOURCE_CACHE_TTL = 60
PUBLISHED_KEY = "x-request-published"
TTL_KEY = "x-request-ttl"
REQUEST_ID_KEY = "x-request-id"
OPAQUE_TAG_KEY = "x-dynatrace"
URI_KEY = "uri"
RECEIPT_HANDLE_KEY = "receipt_handle"
AWS_SNS = "SNS"
AWS_SQS = "SQS"



"""
Consume Daemon main.

"""
from microcosm.loaders import load_each, load_from_dict
from microcosm_daemon.api import SleepNow
from microcosm_daemon.daemon import Daemon

from microcosm_pubsub.consumer import STDIN
from microcosm_pubsub.envelope import LambdaSQSEnvelope, NaiveSQSEnvelope, SQSEnvelope


class ConsumerDaemon(Daemon):

    def __init__(self, event=None):
        super().__init__()
        self.bound_handlers = None
        self.sqs_event = event or {}

    def make_arg_parser(self):
        parser = super().make_arg_parser()
        parser.add_argument("--stdin", action="store_true")
        parser.add_argument("--sqs-queue-url")
        parser.add_argument("--envelope", choices=[
            envelope_cls.__name__
            for envelope_cls in SQSEnvelope.__subclasses__()
        ])
        return parser

    def create_object_graph_components(self, graph):
        super().create_object_graph_components(graph)
        self.bound_handlers = graph.sqs_message_handler_registry.compute_bound_handlers(
            self.components,
        )

    def run_state_machine(self):
        for media_type, handler in self.bound_handlers.items():
            self.graph.logger.info("Handling: {} with handler: {}".format(
                media_type,
                handler.__name__,
            ))

        super().run_state_machine()

    def process(self):
        """
        Lambda Function method that runs only once
        """
        self.initialize()
        self.graph.logger.info("Local starting daemon {}".format(self.name))
        with self.graph.error_policy:
            self.graph.sqs_message_dispatcher.handle_batch(self.bound_handlers)

    @classmethod
    def make_lambda_handler(cls):
        def handler(event, context):
            """
            AWS Lambda function handler.
            """
            # this is for the warmup event.
            # just return something and don't continue
            if "warm" in event:
                return "warming up"

            # we configure SQS Queue to use batches of 1,
            # so received event contains
            # another stringified json with actual message inside
            daemon = cls(event=event["Records"][0])
            daemon.process()
        return handler

    @property
    def defaults(self):
        config = dict()

        if self.args.stdin:
            config.update(
                sqs_envelope=dict(
                    strategy_name=NaiveSQSEnvelope.__name__,
                ),
                sqs_consumer=dict(
                    sqs_queue_url=STDIN,
                    sqs_event=""
                ),
            )

        if self.args.sqs_queue_url:
            config.update(
                sqs_consumer=dict(
                    sqs_queue_url=self.args.sqs_queue_url,
                    sqs_event=""
                ),
            )

        if self.args.envelope:
            config.update(
                sqs_envelope=dict(
                    strategy_name=self.args.envelope,
                ),
            )
        # for AWS Lambda
        # SQS message comes as function argument
        # e.g
        # ```
        # def handler(event, context):
        #     # processing event
        # ```
        # we don't use command line args here
        if self.sqs_event:
            config.update(
                sqs_envelope=dict(
                    strategy_name=LambdaSQSEnvelope.__name__,
                ),
                sqs_consumer=dict(
                    sqs_event=self.sqs_event,
                    sqs_queue_url="",
                )
            )
        return config

    @property
    def components(self):
        return super().components + [
            "opaque",
            "pubsub_message_schema_registry",
            "sqs_message_handler_registry",
            "sqs_consumer",
            "sqs_message_dispatcher",
        ]

    def __call__(self, graph):
        """
        Implement daemon by sinking messages from the consumer to a dispatcher function.

        """
        results = graph.sqs_message_dispatcher.handle_batch(self.bound_handlers)
        if not results:
            raise SleepNow()

    @classmethod
    def create_for_testing(cls, loader=None, **kwargs):
        mock_config = load_from_dict(
            sns_producer=dict(
                mock_sns=False,
            ),
        )

        if loader is None:
            loader = mock_config
        else:
            loader = load_each(loader, mock_config)

        return super().create_for_testing(
            sqs_queue_url="queue",
            loader=loader,
            envelope=None,
            stdin=False,
            **kwargs
        )



"""
Message encoding and decoding.

"""
from json import dumps, loads
from typing import Any, Dict, Optional

from marshmallow import (
    EXCLUDE,
    Schema,
    ValidationError,
    fields,
)


DEFAULT_MEDIA_TYPE = "application/json"


def enrich(error: ValidationError, schema: Schema):
    """
    Enrich a schema validation error with a human-interpretable
    description of a marshmallow schema.

    """
    messages = error.messages
    if isinstance(messages, dict):
        messages.update(dict(
            schema=str(schema),
        ))
    else:
        messages.append(
            str(schema),
        )

    return error


class PubSubMessageSchema(Schema):
    """
    Base schema for messages, including a media type.

    """
    MEDIA_TYPE = DEFAULT_MEDIA_TYPE

    mediaType = fields.String(
        attribute="media_type",
        required=True,
    )
    opaqueData = fields.Dict(
        attribute="opaque_data",
        required=False,
    )


class MediaTypeSchema(Schema):
    """
    Custom schema for extracting media type.

    """
    mediaType = fields.String(required=True)


class PubSubMessageCodec:
    """
    Message encoder/decoder.

    """

    def __init__(self, schema: PubSubMessageSchema):
        self.schema = schema

    def encode(self, dct: Optional[Dict[str, Any]] = None, **kwargs) -> str:
        """
        Encode a message.

        Uses the appropriate codec to write JSON.

        """
        message = dct.copy() if dct else dict()
        message.update(kwargs)

        dumped = self.schema.dump(message)

        # NB: Schema dump doesn't run validation
        errors = self.schema.validate(dumped)
        if errors:
            raise ValidationError(message=errors)

        # Dump to string
        return dumps(dumped)

    def decode(self, message):
        """
        Decode a message.

        Uses the appropriate codec to read JSON.

        """
        if not isinstance(message, dict):
            dct = loads(message)
        else:
            dct = message

        try:
            # load performs a validation and raises on error
            # Similarly, we exclude unknown fields to avoid errors on decode
            return self.schema.load(dct, unknown=EXCLUDE)
        except ValidationError as error:
            # Add more useful information to logs for debugging validation errors.
            raise enrich(error, self.schema)



class AttributeNotFound(Exception):
    def __init__(self, parent, attribute):
        super().__init__(f"Failed to find attribute `{attribute}` on `{parent}`")


class ContextKeyNotFound(Exception):
    def __init__(self, key_error: KeyError, func):
        context_key, *_ = key_error.args
        super().__init__(f"Failed to find context key `{context_key}` during evaluation of `{func}`")



"""
Consumer backoff policy

"""
from abc import ABCMeta, abstractmethod
from random import randint


# We cannot go over 12 hours
MAX_BACKOFF_TIMEOUT = 60 * 60 * 12


class BackoffPolicy(metaclass=ABCMeta):

    @abstractmethod
    def compute_backoff_timeout(self, message, message_timeout):
        pass

    @classmethod
    def choose_backoff_policy(cls, name):
        for subclass in cls.__subclasses__():
            if subclass.__name__ == name:
                return subclass
        raise Exception("No backoff policy configured with class name: {}".format(name))


class NaiveBackoffPolicy(BackoffPolicy):
    """
    Naive backoff policy.

    Uses a fixed timeout from the message or system default.

    """
    def __init__(self, message_retry_visibility_timeout_seconds, **kwargs):
        self.message_retry_visibility_timeout_seconds = message_retry_visibility_timeout_seconds

    def compute_backoff_timeout(self, message, message_timeout):
        backoff_timeout = message_timeout or self.message_retry_visibility_timeout_seconds
        # we can only set integer timeouts
        return int(backoff_timeout) if backoff_timeout is not None else backoff_timeout


class ExponentialBackoffPolicy(BackoffPolicy):
    """
    Exponential backoff policy.

    Uses a timeout scaled between 1 and an exponential limit.

    """
    def __init__(self, **kwargs):
        pass

    def compute_backoff_timeout(self, message, message_timeout):
        # exponential backoff means that on the Cth failure, timeout is maximized at N=2^C - 1
        upper = 2**message.approximate_receive_count - 1

        # randomly select a timeout between 1..N; note that proper exponential backoff uses 0..N
        scaling_factor = self.randint(1, upper)

        return min(scaling_factor, MAX_BACKOFF_TIMEOUT)

    def randint(self, lower, upper):
        """
        Test friendly randomization.

        """
        return randint(lower, upper)


class ExponentialBackoffBaseJitterPolicy(ExponentialBackoffPolicy):
    """
    Exponential backoff jitter with base policy.

    Uses a timeout scaled between number or retry and an exponential limit.

    """
    def __init__(self, **kwargs):
        pass

    def compute_backoff_timeout(self, message, message_timeout):
        # Slow down exponentially, but add a random equal jitter to avoid thundering herd
        # we use the base as the number of times the message has been received (1 second for each minimum wait)
        # the time we wait is between half the base + random number between 0 and half the scaling factor
        # this means we wait:
        #   0-1 second on the original message (this never happens),
        #   1-2 second for the first retry,
        #   2-4 seconds for the second retry,
        #   3-7 seconds for the third retry,
        #   4-12 seconds for the fourth retry,
        #   5-21 seconds for the fifth retry, etc.
        base = message.approximate_receive_count
        upper = int(2**base)
        scaling_factor = self.randint(base, upper)

        return min(int(base + self.randint(0, int(scaling_factor/2))), MAX_BACKOFF_TIMEOUT)



"""
Process batches of messages.

Flow diagram:

+-----------------------+                        +-------------+                                            +----------+
|                       |                        |             |                                            |          |
|                       |     1. Get messages    |             |                                            |          |
| SQSMessageDispatcher  |<-----------------------+             +<-------------------------------------------+          |
|                       |                        |             |                                            |          |
|                       |                        |             |                                            |          |
+-------+-+-------------+                        |             |                                            |          |
           | ˆ                                   |             |                                            |          |
           | | 2. Get handling result            |             |                                            |          |
           | |        +-------------------+      |             |                                            |          |
           | |        |                   |      | SQSConsumer |                                            |   SQS    |
           | |        |                   |      |             |                                            |          |
           | +------->+     Handlers      |      |             |                                            |          |
           |          |                   |      |             |                                            |          |
           |          |                   |      |             |                                            |          |
           |          +-------------------+      |             |                                            |          |
           |                                     |             |                                            |          |
           |  3. Propagate result to SQS         |             |  Action taken depends on handler result:   |          |
           +------------------------------------>+             +------------------------------------------->+          |
                                                 |             |  - SUCCEED/SKIP: delete message from queue |          |
                                                 +-------------+  - FAIL/RETRY: change message visibility   +----------+
                                                                    for future reprocessing.

"""
from logging import Logger
from time import time
from typing import List

from inflection import titleize
from microcosm.api import defaults, typed
from microcosm_logging.decorators import context_logger, logger
from microcosm_logging.timing import elapsed_time

from microcosm_pubsub.constants import PUBLISHED_KEY, TTL_KEY
from microcosm_pubsub.errors import IgnoreMessage, SkipMessage, TTLExpired
from microcosm_pubsub.result import MessageHandlingResult, MessageHandlingResultType
from microcosm_pubsub.tracing import trace_incoming_message_process


@logger
@defaults(
    # Number of failed attempts after which the message stops being processed
    message_max_processing_attempts=typed(int, default_value=None),
)
class SQSMessageDispatcher:
    """
    Dispatch batches of SQSMessages to handler functions.

    """
    logger: Logger

    def __init__(self, graph):
        self.opaque = graph.opaque
        self.sqs_consumer = graph.sqs_consumer
        self.sqs_message_context = graph.sqs_message_context
        self.sqs_message_handler_registry = graph.sqs_message_handler_registry
        self.send_metrics = graph.pubsub_send_metrics
        self.send_batch_metrics = graph.pubsub_send_batch_metrics
        self.max_processing_attempts = graph.config.sqs_message_dispatcher.message_max_processing_attempts
        self.sentry_config = graph.sentry_logging_pubsub

    def handle_batch(self, bound_handlers) -> List[MessageHandlingResult]:
        """
        Send a batch of messages to a function.

        """
        start_time = time()

        instances = [
            self.handle_message(message, bound_handlers)
            for message in self.sqs_consumer.consume()
        ]

        batch_elapsed_time = (time() - start_time) * 1000

        message_batch_size = len([
            instance for instance in instances
            if instance.result != MessageHandlingResultType.IGNORED
        ])

        if message_batch_size > 0:
            # NB: Expose formatted message
            message = "Completed batch: Message count: {message_batch_size}, elapsed_time: {batch_elapsed_time}".format(
                message_batch_size=message_batch_size,
                batch_elapsed_time=batch_elapsed_time,
            )
            self.logger.debug(message)

        self.send_batch_metrics(batch_elapsed_time, message_batch_size)

        for instance in instances:
            self.send_metrics(instance)

        return instances

    def handle_message(self, message, bound_handlers) -> MessageHandlingResult:
        """
        Handle a message.

        """
        with self.opaque.initialize(self.sqs_message_context, message):
            handler = None
            start_handle_time = time()

            queue_url = self.sqs_consumer.sqs_queue_url
            with trace_incoming_message_process(self.opaque, message, queue_url), \
                    elapsed_time(self.opaque):
                try:
                    self.validate_message(message)
                    handler = self.find_handler(message, bound_handlers)
                    instance = MessageHandlingResult.invoke(
                        handler=self.wrap_handler(handler),
                        message=message,
                    )
                except Exception as error:
                    instance = MessageHandlingResult.from_error(
                        message=message,
                        error=error,
                    )

            instance.elapsed_time = self.opaque["elapsed_time"]
            published_time = self.opaque.get(PUBLISHED_KEY)
            if published_time:
                instance.handle_start_time = start_handle_time - float(published_time)
            instance.log(
                logger=self.choose_logger(handler),
                opaque=self.opaque,
            )
            instance.error_reporting(
                sentry_config=self.sentry_config,
                opaque=self.opaque,
            )
            instance.resolve(message)
            return instance

    def validate_message(self, message):
        self.validate_ttl()
        self.validate_processing_limit(message)
        self.validate_content(message)

    def validate_ttl(self):
        """
        Validate that the current message is not expired.

        :raises TTLExpired
        """
        if TTL_KEY not in self.opaque:
            return

        if int(self.opaque[TTL_KEY]) <= 0:
            raise TTLExpired()

    def validate_processing_limit(self, message):
        if self.max_processing_attempts and message.approximate_receive_count > self.max_processing_attempts:
            raise SkipMessage(
                "Message exceeded maximum of {max} processing attempts. Skipping",
                extra=dict(max=self.max_processing_attempts),
            )

    def validate_content(self, message):
        """
        Extract message content.

        The `CodecMediaTypeAndContentParser` (used by the default `CodecSQSEnvelope`) will return
        `None` for content if it doesn't have a registered media type schema. In most cases, this
        condition will coincide with not having a registerd handler as well. Plus, we can't handle
        absent content.

        """
        if message.content is None:
            raise IgnoreMessage(f"Could not parse message for: {message.media_type}")

    def find_handler(self, message, bound_handlers):
        """
        Find the handler for a message.

        """
        try:
            handler = self.sqs_message_handler_registry.find(message.media_type, bound_handlers)
            self.opaque["handler"] = titleize(handler.__class__.__name__)
            return handler
        except KeyError:
            raise IgnoreMessage(f"No handler was registered for: {message.media_type}")

    def wrap_handler(self, handler):
        """
        Wrap handler with context logger.

        Ensures that all handler logger calls have access to opaque data.

        """
        return context_logger(
            context_func=lambda *args, **kwargs: self.opaque,
            func=handler,
            parent=handler,
        )

    def choose_logger(self, handler):
        """
        Choose a logger to use for handler results.

        """
        try:
            # use the handler's logger, if possible
            return handler.logger
        except AttributeError:
            return self.logger



"""
Hamcrest-style matchers for PubSub messages.

Note that this package depends on `PyHamcrest`, which is intentionally not an
installation dependency of `microcosm-pubsub` (but is a test dependnency).

Users are expected to install `PyHamcrest` as a test dependency and import
this package as part of their unit tests (and not their runtime).

"""
from microcosm_pubsub.matchers.message import (
    PublishedMessage,
    PublishedMessageMatcher,
    has_media_type,
    has_uri,
)
from microcosm_pubsub.matchers.publishing import (
    published,
    published_inanyorder,
    published_messages_for,
    published_nothing,
)


__all__ = [  # type: ignore
    has_media_type,
    has_uri,
    published,
    published_inanyorder,
    published_messages_for,
    published_nothing,
    PublishedMessage,
    PublishedMessageMatcher,
]



"""
Support optional tracing using Dynatrace oneagent.

If the dynatrace SDK is not installed, does nothing.

"""
import json
from contextlib import contextmanager

from microcosm.opaque import Opaque
from microcosm_pubsub.constants import AWS_SNS, AWS_SQS, OPAQUE_TAG_KEY, REQUEST_ID_KEY

from microcosm_pubsub.message import SQSMessage


try:
    import oneagent
    from oneagent.common import ChannelType, MessagingDestinationType
    from oneagent.sdk import Channel
except ImportError:
    oneagent = None
    ChannelType = None
    MessagingDestinationType = None
    Channel = None


class StubOutgoingTracer:
    outgoing_dynatrace_string_tag = None

    def set_vendor_message_id(self, id):
        pass

    def set_correlation_id(self, id):
        pass


def add_trace_to_message(tracer, pubsub_message):
    tag = tracer.outgoing_dynatrace_string_tag
    if tag:
        try:
            message = json.loads(pubsub_message.message)
            message["opaqueData"][OPAQUE_TAG_KEY] = tag
            pubsub_message.message = json.dumps(message)
        except Exception:
            pass


@contextmanager
def trace_outgoing_message(topic_arn: str):
    if not oneagent:
        yield StubOutgoingTracer()
    else:
        sdk = oneagent.get_sdk()
        try:
            topic_name = topic_arn.split(":")[-1]
        except IndexError:
            topic_name = topic_arn

        messaging_system = sdk.create_messaging_system_info(
            AWS_SNS,
            topic_name,
            MessagingDestinationType.TOPIC,
            Channel(ChannelType.TCP_IP, None),
        )
        with sdk.trace_outgoing_message(messaging_system) as tracer:
            yield tracer


@contextmanager
def trace_incoming_message_process(opaque: Opaque, message: SQSMessage, queue_url: str):
    if not oneagent:
        yield
    else:
        try:
            queue_name = queue_url.split("/")[-1]
        except IndexError:
            queue_name = queue_url

        sdk = oneagent.get_sdk()
        messaging_system = sdk.create_messaging_system_info(
            AWS_SQS,
            queue_name,
            MessagingDestinationType.QUEUE,
            Channel(ChannelType.TCP_IP, None),
        )

        tag = opaque.get(OPAQUE_TAG_KEY)
        request_id = opaque.get(REQUEST_ID_KEY)
        with sdk.trace_incoming_message_process(messaging_system, str_tag=tag) as tracer:
            tracer.set_vendor_message_id(message.message_id)
            if request_id:
                tracer.set_correlation_id(request_id)
            yield



"""
for_each("item").in_("items").do(...)

"""
from microcosm_pubsub.chain import Chain


def yes(value):
    return True


def not_none(value):
    return value is not None


class ForEachStatement:
    def __init__(self, key, list_key=None):
        self.key = key
        self.items = None
        self.chain = None
        self.list_key = list_key or f"{self.key}_list"
        self.filter_func = yes

    def __str__(self):
        return f"for_{self.key}"

    def in_(self, items):
        self.items = items
        return self

    def as_(self, list_key):
        self.list_key = list_key
        return self

    def when(self, filter_func):
        self.filter_func = filter_func
        return self

    def when_not_none(self):
        return self.when(not_none)

    def do(self, *args, **kwargs):
        self.chain = Chain.make(*args, **kwargs)
        return self

    def __call__(self, context):
        values = [
            self.chain(context.local(**{
                self.key: item,
            }))
            for item in context[self.items]
        ]

        filtered_values = list(filter(self.filter_func, values))

        # Set the responses in the context
        context[self.list_key] = filtered_values
        return filtered_values


def for_each(key):
    """
    Run the chain for all items in array
    Each items will be accessible as "item" name

    Example: for("item")
        .in_(items)
        .do(Chain(...))

    """
    return ForEachStatement(key)
